# Copyright 2023 Eurobios
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

from abc import ABCMeta, abstractmethod, abstractproperty
from typing import Union, List

from palma.base.model_selection import ModelSelector
from palma.base.project import Project


class Component(object, metaclass=ABCMeta):

    def __str__(self):
        return self.__class__.__name__


class ProjectComponent(Component):
    """
    Base Project Component class

    This object ensures that all subclasses Project component implements a
    """

    @abstractmethod
    def __call__(self, obj) -> None:
        ...


class ModelComponent(Component):
    """
    Base Model Component class
    """

    def __call__(self, project, model):
        pass


class Logger(Component, metaclass=ABCMeta):
    """
    Logger is an abstract class that defines a common
    interface for a set of Logger-subclasses.

    It provides common methods for all possible subclasses, making it 
    possible for a user to create a custom subclass compatible  with 
    the rest of the components. 
    """

    @abstractmethod
    def __init__(self, **kwargs) -> None:
        pass

    def __call__(self, obj: Union["Project", "ModelSelector"]) -> None:
        if isinstance(obj, Project):
            self.log_project(obj)

        elif isinstance(obj, ModelSelector):
            self.log_run(obj)

    @abstractmethod
    def log_project(self, project: 'Project') -> None:
        """
        Saves an instance of :class:`~autolm.project.Project`
        """
        ...

    @abstractmethod
    def log_run(self, run: 'ModelSelector') -> None:
        """
        Saves all relevant items and artefacts
        generated by an instance of :class:`~autolm.run.Run`. \
        This abstract method should call self._log_metrics, 
        self._log_params, self._log_model. 
        """
        ...

    @abstractmethod
    def _log_metrics(self, **kwargs) -> None:
        """
        Saves the metrics of an instance of \ 
        :class:`~autolm.run.Run` using its attributes.
        Example :
            :attr:`~autolm.run.Run.metrics`
        """
        ...

    @abstractmethod
    def _log_params(self, **kwargs) -> None:
        """
        Saves the hyperOpt parameters of an instance of \ 
        :class:`~autolm.run.Run` using its attributes and/or 
        methods. 
        Example :
            :meth:`autolm.run.Run.model.get_params()`
        """
        ...

    @abstractmethod
    def _log_model(self, **kwargs) -> None:
        """
        Saves the model of an instance of \ 
        :class:`~autolm.run.Run` using its attributes.
        Example :
            :attr:`~autolm.run.Run.model`
        """
        ...
